/*
 * Copyright (c) 1994
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* $Id: if.c,v 1.3 1994/12/06 00:19:51 he Exp $ */

#include <sys/types.h>
#include <sys/time.h>

#include <stdio.h>
#include <signal.h>

#include <isode/pepsy/SNMP-types.h>
#include <isode/snmp/objects.h>
#include <isode/tailor.h>
#include <isode/dgram.h>
#include <isode/tsap.h>
#include <isode/internet.h>

#include "common.h"
#include "snmp_aux.h"
#include "utils.h"
#include "if.h"

extern void bcopy ();
extern int printf ();
extern int fflush ();
extern int gettimeofday ();
extern u_int alarm ();
extern int sigblock ();
extern int sigpause ();
extern int sigsetmask ();


typedef struct ip_addr_entry* IPE;

struct ip_addr_entry {
    struct sockaddr_in *address;
    struct sockaddr_in *mask;
    int index;
    IPE next;
};

static void catchalarm();

static OID oid_ipAdEntIfIndex;
static OID oid_ipAdEntAddr;
static OID oid_ipAdEntNetMask;
static int signalled;
static IPE ip_addr_info = NULL;


void
process_ip_addr_info (vbl)
    struct type_SNMP_VarBindList *vbl;
{
    struct type_SNMP_VarBindList *vp;
    struct type_SNMP_VarBind *v;
    caddr_t value;
    OI oi;
    OS os;
    IPE ip_info, tp;

    ip_info = (IPE) calloc(sizeof *ip_info, 1);
    if (ip_addr_info) 
    {
	for(tp = ip_addr_info; tp -> next; tp = tp -> next);
	tp -> next = ip_info;
    }
    else
	ip_addr_info = ip_info;

    ip_info -> address =
	(struct sockaddr_in*) calloc(sizeof(struct sockaddr_in), 1);
    ip_info -> mask =
	(struct sockaddr_in*) calloc(sizeof(struct sockaddr_in), 1);

    for (vp = vbl; vp; vp = vp -> next) {
	v = vp -> VarBind;

	if ((oi = name2inst (v -> name)) == NULL ||
	    (os = oi -> oi_type -> ot_syntax) == NULL ||
	    (*os -> os_decode) (&value, v -> value) == NOTOK)
	{
	    vunknown (v -> value);
	}
	else
	{
	    /* Layering violations, I know... */
	    if (oid_pfx_match (oid_ipAdEntIfIndex, v -> name)) 
		ip_info -> index = *((int*) value);
	    if (oid_pfx_match (oid_ipAdEntAddr, v -> name))
		bcopy ((char*) value, (char*) ip_info -> address,
		       sizeof (struct sockaddr_in));
	    if (oid_pfx_match (oid_ipAdEntNetMask, v -> name))
		bcopy ((char*) value, (char*) ip_info -> mask,
		       sizeof (struct sockaddr_in));

	    (*os -> os_free) (value);
	}
    }    
}


void
get_ip_addr_info (ps, sd)
    PS ps;
    int sd;
{
    struct type_SNMP_Message *msg;

    msg = new_message (type_SNMP_PDUs_get__next__request, community);
    if (!msg)
	adios (NULLCP, "out of memory");
    
    oid_ipAdEntIfIndex = add_null_varbind (msg, "ipAdEntIfIndex");
    oid_ipAdEntAddr    = add_null_varbind (msg, "ipAdEntAddr");
    oid_ipAdEntNetMask = add_null_varbind (msg, "ipAdEntNetMask");
    
    getnext_process (msg, ps, sd, process_ip_addr_info);
}


IPE
find_ip_info (ix)
    int ix;
{
    IPE ip;

    for (ip = ip_addr_info; ip; ip = ip -> next) 
    {
	if (ip -> index == ix)
	    return ip;
    }
    return NULL;
}


void
print_if_descr (fmt, value)
    char *fmt;
    caddr_t value;
{
    struct qbuf *qb;
    struct qbuf *x = (struct qbuf*) value;
    int len = 0;
    char *buf, *p = NULL;

    if (!x)
    {
	printf (fmt, "-unknown-");
	return;
    }

    for (qb = x -> qb_forw; qb != x; qb = qb -> qb_forw)
	len += qb -> qb_len;
    buf = calloc(len+1, 1);
    
    for (qb = x -> qb_forw, p = buf; qb != x; qb = qb -> qb_forw)
    {
	bcopy(qb -> qb_data, p, qb -> qb_len);
	p += qb -> qb_len;
    }

    /* Fix up so we get only one field (insert hyphens) */
    for (p = buf; p < &buf[len]; p++)
	if (*p == ' ')
	    *p = '-';

    buf[len] = 0;
    printf (fmt, buf);
    free (buf);
}


static char *admin_status[] = {
    "on", "off", "tst"
};

void
print_admin_status(fmt, value)
    char *fmt;
    caddr_t value;
{

    enum_print(fmt, value, admin_status,
	       sizeof admin_status / sizeof admin_status[0]);
}
    

static char *oper_status[] = {
    "up", "down", "test"
};

void
print_oper_status(fmt, value)
    char *fmt;
    caddr_t value;
{

    enum_print(fmt, value, oper_status,
	       sizeof oper_status / sizeof oper_status[0]);
}


static char *if_types[] = {
    "other", "1822", "hdh-1822", "ddn-x25", "877-x25", "ether", "802.3",
    "tk-bus", "tk-ring", "802.6-man", "starlan", "prot-10", "prot-80",
    "hyperch", "fddi", "lapb", "sdlc", "t1", "e1", "bas-isdn", "pri-isdn",
    "prop-ser", "ppp", "loopback", "eon", "3Mb-ether", "nsip", "slip",
    "ultra", "ds3", "sip", "frame-rel", "rs232", "para", "arcnet",
    "arcnet+", "atm", "miox25", "sonet", "x25ple", "88022llc", "localtalk",
    "smdsdxi", "fr-dce", "v35", "hssi", "hippi", "modem", "aal5-atm",
    "sonetpath", "sonet-vt", "smds-icip", "prop-virt", "prop-mux"
};

void
print_if_type(fmt, value)
    char *fmt;
    caddr_t value;
{

    enum_print(fmt, value, if_types,
	       sizeof if_types / sizeof if_types[0]);
}

void
print_int (fmt, value)
    char *fmt;
    caddr_t value;
{
    char buf[20];

    if (!value)
    {
	printf (fmt, "**");
	return;
    }

    sprintf (buf, "%lu", *((u_long*) value));
    printf (fmt, buf);
}

void
print_mtu(fmt, value)
    char *fmt;
    caddr_t value;
{

    print_int (fmt, value);
}

void
print_media_addr(fmt, value)
    char *fmt;
    caddr_t value;
{
    char *buf, buf1[5];
    char *p, *cp, *ep;
    struct qbuf *qb, *x = (struct qbuf*) value;
    int len;

    for (len = 0, qb = x -> qb_forw; qb != x; qb = qb -> qb_forw)
	len += qb -> qb_len;

    if (len == 0) 
    {
	printf (fmt, "(null)");
	return;
    }
    
    len *= 3;
    buf = calloc (len, 1);

    p = "";
    for (qb = x -> qb_forw; qb != x; qb = qb -> qb_forw)
    {
	for (ep = (cp = qb -> qb_data) + qb -> qb_len; cp < ep; cp++) 
	{
	    sprintf (buf1, "%s%x", p, *cp & 0xff);
	    strcat (buf, buf1);
	    p = ":";
	}
    }
    printf (fmt, buf);
    free (buf);
}


char *
sprint_pretty_value (rate)
    double rate;
{
    static char buf[20];

    if ( rate < 0 )
	return "neg? ";
    if ( rate < 1.0e-6 )
	return "0u";
    if ( rate < 1.0e-3 )
    {
	sprintf (buf, "%3du", (int) (rate*1e6));
	return buf;
    }
    if ( rate < 1 && rate > 0.1 )
    {
	sprintf (buf, "%3.1f ", rate);
	return buf;
    }
    if ( rate < 1 )
    {
	sprintf (buf, "%3dm", (int) (rate*1e3));
	return buf;
    }
    if ( rate < 10 )
    {
	sprintf (buf, "%3.1f ", rate);
	return buf;
    }
    if ( rate < 1000 )
    {
	sprintf (buf, "%3d ", (int) rate);
	return buf;
    }
    if ( rate < 10000 ) 
    {
	sprintf (buf, "%3.1fk", rate / 1000);
	return buf;
    }
    if ( rate < 1e6 )
    {
	sprintf (buf, "%3dk", (int) (rate / 1000));
	return buf;
    }
    if ( rate < 1e7 )
    {
	sprintf (buf, "%3.1fM", rate / 1e6);
	return buf;
    }
    if ( rate < 1e9 )
    {
	sprintf (buf, "%3dM", (int) (rate / 1e6));
	return buf;
    }
    if ( rate < 1e10 )
    {
	sprintf (buf, "%3.1fG", rate / 1e9);
	return buf;
    }
    if ( rate < 1e12 )
    {
	sprintf (buf, "%3dG", (int) (rate / 1e9));
	return buf;
    }
    if ( rate < 1e13 )
    {
	sprintf (buf, "%3.1fT", rate / 1e12);
	return buf;
    }
    return ">10T";
}    

double
time_diff(newstamp, oldstamp)
    struct timeval *newstamp, *oldstamp;
{
    double diff;
    
    diff = newstamp->tv_sec - oldstamp->tv_sec;
    diff += (newstamp->tv_usec - oldstamp->tv_usec) / 100000.0;
    return diff;
}

char *
sprint_pretty_rate (newval, oldval, newstamp, oldstamp)
    int newval, oldval;
    struct timeval *newstamp, *oldstamp;
{
    int diff;
    double seconds, rate;

    if (newval == oldval)
	return "0 ";
    diff = newval - oldval;
    seconds = time_diff(newstamp, oldstamp);
    rate = ((double) diff) / seconds;
    return sprint_pretty_value (rate);
}

char *
sprint_pretty_bitrate (newbytes, oldbytes, newstamp, oldstamp)
    int newbytes, oldbytes;
    struct timeval *newstamp, *oldstamp;
{
    int diff;
    double seconds, rate;

    if (newbytes == oldbytes)
	return "0 ";
    diff = newbytes - oldbytes;
    seconds = time_diff(newstamp, oldstamp);
    rate = ((double) diff) * 8.0 / seconds;
    return sprint_pretty_value (rate);
}

void
print_speed (fmt, value)
    char *fmt;
    caddr_t value;
{
    int speed = *((int*) value);
    
    printf (fmt, sprint_pretty_value ((double) speed));
}
    

static OID oid_ifIndex;
static OID oid_ifDescr;
static OID oid_ifPhysAddress;
static OID oid_ifMtu;
static OID oid_ifAdminStatus;
static OID oid_ifOperStatus;
static OID oid_ifSpeed;
static OID oid_ifType;
static OID oid_ifInOctets;
static OID oid_ifInUcastPkts;
static OID oid_ifInNUcastPkts;
static OID oid_ifInErrors;
static OID oid_ifInDiscards;
static OID oid_ifOutOctets;
static OID oid_ifOutUcastPkts;
static OID oid_ifOutNUcastPkts;
static OID oid_ifOutErrors;
static OID oid_ifOutDiscards;
static OID oid_ifOutQLen;

struct var_info {
    caddr_t value;
    OS os;
    struct var_info *next;
};


void
process_if_info (vbl)
    struct type_SNMP_VarBindList *vbl;
{
    struct type_SNMP_VarBindList *vp;
    struct type_SNMP_VarBind *v;
    caddr_t value;
    OI oi;
    OS os;
    caddr_t ifIndex, ifDescr, ifPhysAddress, ifMtu;
    caddr_t ifAdminStatus, ifOperStatus, ifSpeed, ifType;
    IPE ip_info;
    struct var_info *vars = NULL;
    struct var_info *new_var, *vip;
    int ix;

    ifIndex = ifDescr = ifPhysAddress = ifMtu = ifAdminStatus =
	ifOperStatus = ifSpeed = ifType = NULL;

    for (vp = vbl; vp; vp = vp -> next) {
	v = vp -> VarBind;

	if ((oi = name2inst (v -> name)) == NULL ||
	    (os = oi -> oi_type -> ot_syntax) == NULL ||
	    (*os -> os_decode) (&value, v -> value) == NOTOK)
	{
	    vunknown (v -> value);
	}
	else
	{
	    new_var = (struct var_info*) calloc(sizeof (struct var_info), 1);
	    new_var -> next = vars;
	    vars = new_var;

	    new_var -> value = value;
	    new_var -> os = os;

	    if (oid_pfx_match (oid_ifIndex, v -> name))
		ifIndex = value;
	    if (oid_pfx_match (oid_ifDescr, v -> name))
		ifDescr = value;
	    if (oid_pfx_match (oid_ifAdminStatus, v -> name))
		ifAdminStatus = value;
	    if (oid_pfx_match (oid_ifOperStatus, v -> name))
		ifOperStatus = value;
	    if (oid_pfx_match (oid_ifPhysAddress, v -> name))
		ifPhysAddress = value;
	    if (oid_pfx_match (oid_ifSpeed, v -> name))
		ifSpeed = value;
	    if (oid_pfx_match (oid_ifType, v -> name))
		ifType = value;
	    if (oid_pfx_match (oid_ifMtu, v -> name))
		ifMtu = value;
	}
    }	
    /* Layering violations, I know... */
    ix = *((int*) ifIndex);

    printf ("%-2d ", ix);
    print_if_descr ("%-19.19s ", ifDescr);

    if (!mflag) 
    {
	if (ifAdminStatus)
	    print_admin_status("%3s ", ifAdminStatus);
	else
	    printf ("%3s ", "**");

	if (ifOperStatus)
	    print_oper_status("%4s ", ifOperStatus);
	else
	    printf ("%4s ", "**");
    }

    if ((ip_info = find_ip_info(ix)) == NULL) 
    {
	printf ("%-15.15s ", "none");
	if (!mflag)
	    printf ("%-15.15s ", "none");
    }
    else 
    {
	printf ("%-15.15s ", inet_ntoa(ip_info -> address -> sin_addr));
	if (!mflag)
	    printf ("%-15.15s ", inet_ntoa(ip_info -> mask -> sin_addr));
    }

    if (mflag) 
    {
	print_media_addr("%-17.17s ", ifPhysAddress);
	print_speed("%6s ", ifSpeed);
    }

    if (ifType)
	print_if_type("%9s ", ifType);
    else
	printf ("%9s ", "-unknown-");

    print_mtu("%5s", ifMtu);

    printf ("\n");

    for (vip = vars; vip; vip = vip -> next)
	(*vip -> os -> os_free) (vip -> value);
}
    
void
process_B_info (vbl)
    struct type_SNMP_VarBindList *vbl;
{
    struct type_SNMP_VarBindList *vp;
    struct type_SNMP_VarBind *v;
    caddr_t value;
    OI oi;
    OS os;
    caddr_t ifIndex, ifDescr;
    caddr_t ifInOctets, ifInUcastPkts, ifInNUcastPkts;
    caddr_t ifOutOctets, ifOutUcastPkts, ifOutNUcastPkts;
    IPE ip_info;
    struct var_info *vars = NULL;
    struct var_info *new_var, *vip;
    int ix;
    u_long cnt;

    ifIndex = ifDescr =
	ifInOctets = ifInUcastPkts = ifInNUcastPkts =
	ifOutOctets = ifOutUcastPkts = ifOutNUcastPkts = NULL;

    for (vp = vbl; vp; vp = vp -> next) {
	v = vp -> VarBind;

	if ((oi = name2inst (v -> name)) == NULL ||
	    (os = oi -> oi_type -> ot_syntax) == NULL ||
	    (*os -> os_decode) (&value, v -> value) == NOTOK)
	{
	    vunknown (v -> value);
	}
	else
	{
	    new_var = (struct var_info*) calloc(sizeof (struct var_info), 1);
	    new_var -> next = vars;
	    vars = new_var;

	    new_var -> value = value;
	    new_var -> os = os;

	    if (oid_pfx_match (oid_ifIndex, v -> name))
		ifIndex = value;
	    if (oid_pfx_match (oid_ifDescr, v -> name))
		ifDescr = value;
	    if (oid_pfx_match (oid_ifInOctets, v -> name))
		ifInOctets = value;
	    if ( oid_pfx_match (oid_ifInUcastPkts, v -> name))
		ifInUcastPkts = value;
	    if ( oid_pfx_match (oid_ifInNUcastPkts, v -> name))
		ifInNUcastPkts = value;
	    if ( oid_pfx_match (oid_ifOutOctets, v -> name))
		ifOutOctets = value;
	    if ( oid_pfx_match (oid_ifOutUcastPkts, v -> name))
		ifOutUcastPkts = value;
	    if ( oid_pfx_match (oid_ifOutNUcastPkts, v -> name))
		ifOutNUcastPkts = value;
	}
    }	
    /* Layering violations, I know... */
    ix = *((int*) ifIndex);

    printf ("%-2d ", ix);
    print_if_descr ("%-12.12s ", ifDescr);

    if ((ip_info = find_ip_info(ix)) == NULL) 
	printf ("%-15.15s ", "none");
    else 
	printf ("%-15.15s ", inet_ntoa(ip_info -> address -> sin_addr));

    print_int ("%12s ", ifInOctets);

    if (ifInUcastPkts && ifInNUcastPkts)
    {
	cnt =  *((u_long*) ifInUcastPkts);
	cnt += *((u_long*) ifInNUcastPkts);
	printf ("%10lu ", cnt);
    }
    else
	printf ("%10s ", "**");

    print_int ("%12s ", ifOutOctets);

    if (ifOutUcastPkts && ifOutNUcastPkts)
    {
	cnt =  *((u_long*) ifOutUcastPkts);
	cnt += *((u_long*) ifOutNUcastPkts);
	printf ("%10lu", cnt);
    }
    else
	printf ("%10s", "**");

    printf("\n");

    for (vip = vars; vip; vip = vip -> next)
	(*vip -> os -> os_free) (vip -> value);
}


void
process_C_info (vbl)
    struct type_SNMP_VarBindList *vbl;
{
    struct type_SNMP_VarBindList *vp;
    struct type_SNMP_VarBind *v;
    caddr_t value;
    OI oi;
    OS os;
    caddr_t ifIndex, ifDescr;
    caddr_t ifInUcastPkts, ifInNUcastPkts;
    caddr_t ifInErrors, ifInDiscards;
    caddr_t ifOutUcastPkts, ifOutNUcastPkts;
    caddr_t ifOutErrors, ifOutDiscards;
    IPE ip_info;
    struct var_info *vars = NULL;
    struct var_info *new_var, *vip;
    int ix;
    u_long cnt;

    ifIndex = ifDescr = NULL;
    ifInUcastPkts = ifInNUcastPkts = NULL;
    ifOutUcastPkts = ifOutNUcastPkts = NULL;
    ifInErrors = ifInDiscards = ifOutErrors = ifOutDiscards = NULL;
    
    for (vp = vbl; vp; vp = vp -> next) {
	v = vp -> VarBind;

	if ((oi = name2inst (v -> name)) == NULL ||
	    (os = oi -> oi_type -> ot_syntax) == NULL ||
	    (*os -> os_decode) (&value, v -> value) == NOTOK)
	{
	    vunknown (v -> value);
	}
	else
	{
	    new_var = (struct var_info*) calloc(sizeof (struct var_info), 1);
	    new_var -> next = vars;
	    vars = new_var;

	    new_var -> value = value;
	    new_var -> os = os;

	    if (oid_pfx_match (oid_ifIndex, v -> name))
		ifIndex = value;
	    if (oid_pfx_match (oid_ifDescr, v -> name))
		ifDescr = value;
	    if ( oid_pfx_match (oid_ifInUcastPkts, v -> name))
		ifInUcastPkts = value;
	    if ( oid_pfx_match (oid_ifInNUcastPkts, v -> name))
		ifInNUcastPkts = value;
	    if ( oid_pfx_match (oid_ifOutUcastPkts, v -> name))
		ifOutUcastPkts = value;
	    if ( oid_pfx_match (oid_ifOutNUcastPkts, v -> name))
		ifOutNUcastPkts = value;
	    if ( oid_pfx_match (oid_ifInErrors, v -> name))
		ifInErrors = value;
	    if ( oid_pfx_match (oid_ifInDiscards, v -> name))
		ifInDiscards = value;
	    if ( oid_pfx_match (oid_ifOutErrors, v -> name))
		ifOutErrors = value;
	    if ( oid_pfx_match (oid_ifOutDiscards, v -> name))
		ifOutDiscards = value;
	}
    }	
    /* Layering violations, I know... */
    ix = *((int*) ifIndex);

    printf("%-2d ", ix);
    print_if_descr ("%-11.11s ", ifDescr);

    if ((ip_info = find_ip_info(ix)) == NULL) 
	printf ("%-15.15s ", "none");
    else 
	printf ("%-15.15s ", inet_ntoa(ip_info -> address -> sin_addr));

    print_int ("%6s ", ifInErrors);
    print_int ("%6s ", ifInDiscards);
    
    if (ifInUcastPkts && ifInNUcastPkts)
    {
	cnt =  *((u_long*) ifInUcastPkts);
	cnt += *((u_long*) ifInNUcastPkts);
	printf ("%10lu ", cnt);
    }
    else
	printf ("%10s ", "**");

    print_int ("%4s ", ifOutErrors);
    print_int ("%7s ", ifOutDiscards);
    
    if (ifOutUcastPkts && ifOutNUcastPkts)
    {
	cnt =  *((u_long*) ifOutUcastPkts);
	cnt += *((u_long*) ifOutNUcastPkts);
	printf ("%10lu", cnt);
    }
    else
	printf ("%10s", "**");

    printf("\n");

    for (vip = vars; vip; vip = vip -> next)
	(*vip -> os -> os_free) (vip -> value);
}


void
iheading ()
{
    
    printf ("%2s %-19.19s ", "Ix", "Name");
    if (!mflag)
	printf ("%3s %4s ", "as", "os");

    printf ("%-15.15s ", "Address");

    if (!mflag)
	printf ("%-15.15s ", "Netmask");
    else
	printf ("%-17.17s %6s", "Physaddr", "Speed");
	    
    printf ("%9s %5s\n", "Type", "MTU");
}

void
Bheading ()
{

    printf ("%-2s %-12.12s %-15.15s %12s %10s %12s %10s\n",
	    "Ix", "Name", "Address", "ibytes", "ipkts",
	    "obytes", "opkts");
}

void
Bint_heading ()
{

    printf ("%2s %-12.12s %-15.15s %11s %10s %12s %10s\n",
	    "Ix", "Name", "Address", "ibit/s", "ipkts/s",
	    "obit/s", "opkts/s");
}

void
Cheading ()
{

    printf ("%-2s %-11.11s %-15.15s %6s %6s %10s %4s %7s %10s\n",
	    "Ix", "Name", "Address", "ierr", "idisc", "ipkts",
	    "oerr", "odisc", "opkts");
}

void
Cint_heading ()
{

    printf("%2s %-12.12s %-14.14s %6s %8s %6s %6s %7s %8s\n",
	   "Ix", "Name", "Addres", "ierr/s", "ipkts/s", 
	   "oqlen", "oerr/s", "odisc/s", "opkts/s");
}

struct if_counters {
    u_int ipkts, ibytes, ierrs, idisc;
    u_int opkts, obytes, oerrs, odisc, oqlen;
    struct timeval stamp;
};

struct if_counters *old_counters;
struct if_counters *new_counters;

int max_if_num;
struct if_counters *counters;

void (*print_hook) () = NULL;


#define PR(val, ix)	sprint_pretty_rate(new_counters[ix].val,	\
					   old_counters[ix].val,	\
					   &new_counters[ix].stamp,	\
					   &old_counters[ix].stamp)

#define BR(val, ix)	sprint_pretty_bitrate(new_counters[ix].val,	\
					      old_counters[ix].val,	\
					      &new_counters[ix].stamp,	\
					      &old_counters[ix].stamp)

void
print_B_int (ix, descr)
    int ix;
    caddr_t descr;
{
    IPE ip_info;
    
    printf("%-2d ", ix);
    print_if_descr ("%-12.12s ", descr);
    if ((ip_info = find_ip_info(ix)) == NULL) 
	printf ("%-15.15s ", "none");
    else 
	printf ("%-15.15s ", inet_ntoa(ip_info -> address -> sin_addr));

    printf("%12s ", BR(ibytes, ix));
    printf("%10s ", PR(ipkts, ix));
    printf("%12s ", BR(obytes, ix));
    printf("%10s",  PR(opkts, ix));
    printf("\n");
    fflush(stdout);
}

void
print_C_int (ix, descr)
    int ix;
    caddr_t descr;
{
    IPE ip_info;
    
    printf("%-2d ", ix);
    print_if_descr ("%-12.12s ", descr);
    if ((ip_info = find_ip_info(ix)) == NULL) 
	printf ("%-14.14s ", "none");
    else 
	printf ("%-14.14s ", inet_ntoa(ip_info -> address -> sin_addr));
    
    printf("%6s ", PR(ierrs, ix));
    printf("%8s ", PR(ipkts, ix));
    printf("%5d  ", new_counters[ix].oqlen);
    printf("%6s ", PR(oerrs, ix));
    printf("%7s ", PR(odisc, ix));
    printf("%8s", PR(opkts, ix));
    printf("\n");
    fflush(stdout);
}


void
process_int (vbl)
    struct type_SNMP_VarBindList *vbl;
{
    struct type_SNMP_VarBindList *vp;
    struct type_SNMP_VarBind *v;
    caddr_t value;
    OI oi;
    OS os;
    caddr_t ifIndex, ifDescr;
    caddr_t ifInOctets, ifInUcastPkts, ifInNUcastPkts;
    caddr_t ifOutOctets, ifOutUcastPkts, ifOutNUcastPkts;
    caddr_t ifInErrors, ifInDiscards, ifOutErrors, ifOutDiscards, ifOutQLen;
    struct var_info *vars = NULL;
    struct var_info *new_var, *vip;
    int ix = 0;
    int cnt;

    ifIndex = ifDescr = NULL;
    ifInOctets = ifInUcastPkts = ifInNUcastPkts = NULL;
    ifOutOctets = ifOutUcastPkts = ifOutNUcastPkts = NULL;
    ifInErrors = ifInDiscards = ifOutErrors = ifOutDiscards = ifOutQLen =
	NULL;

    for (vp = vbl; vp; vp = vp -> next) {
	v = vp -> VarBind;

	if ((oi = name2inst (v -> name)) == NULL ||
	    (os = oi -> oi_type -> ot_syntax) == NULL ||
	    (*os -> os_decode) (&value, v -> value) == NOTOK)
	{
	    vunknown (v -> value);
	}
	else
	{
	    new_var = (struct var_info*) calloc(sizeof (struct var_info), 1);
	    new_var -> next = vars;
	    vars = new_var;

	    new_var -> value = value;
	    new_var -> os = os;

	    if (oid_pfx_match (oid_ifIndex, v -> name))
		ifIndex = value;
	    if (oid_pfx_match (oid_ifDescr, v -> name))
		ifDescr = value;
	    if (oid_pfx_match (oid_ifInOctets, v -> name))
		ifInOctets = value;
	    if ( oid_pfx_match (oid_ifInUcastPkts, v -> name))
		ifInUcastPkts = value;
	    if ( oid_pfx_match (oid_ifInNUcastPkts, v -> name))
		ifInNUcastPkts = value;
	    if ( oid_pfx_match (oid_ifOutOctets, v -> name))
		ifOutOctets = value;
	    if ( oid_pfx_match (oid_ifOutUcastPkts, v -> name))
		ifOutUcastPkts = value;
	    if ( oid_pfx_match (oid_ifOutNUcastPkts, v -> name))
		ifOutNUcastPkts = value;

	    if ( oid_pfx_match (oid_ifInErrors, v -> name))
		ifInErrors = value;
	    if ( oid_pfx_match (oid_ifInDiscards, v -> name))
		ifInDiscards = value;
	    if ( oid_pfx_match (oid_ifOutErrors, v -> name))
		ifOutErrors = value;
	    if ( oid_pfx_match (oid_ifOutDiscards, v -> name))
		ifOutDiscards = value;
	    if ( oid_pfx_match (oid_ifOutQLen, v -> name))
		ifOutQLen = value;
	}
    }	
    /* Layering violations, I know... */
    if (ifIndex)
	ix = *((int*) ifIndex);
    else
	adios (NULLCP, "didn't return ifIndex");

    if (ix > max_if_num)
	adios (NULLCP, "returned larger than max (%d) ifIndex, %d",
	       max_if_num, ix);

    (void) gettimeofday(&(counters[ix].stamp), NULL);

    if (ifInOctets)
	counters[ix].ibytes = *((int*) ifInOctets);
    if (ifInUcastPkts && ifInNUcastPkts)
    {
	cnt =  *((u_long*) ifInUcastPkts);
	cnt += *((u_long*) ifInNUcastPkts);
	counters[ix].ipkts = cnt;
    }
    if (ifOutOctets)
	counters[ix].obytes = *((int*) ifOutOctets);
    if (ifOutUcastPkts && ifOutNUcastPkts)
    {
	cnt =  *((u_long*) ifOutUcastPkts);
	cnt += *((u_long*) ifOutNUcastPkts);
	counters[ix].opkts = cnt;
    }
    if (ifInErrors)
	counters[ix].ierrs = *((int*) ifInErrors);
    if (ifInDiscards)
	counters[ix].idisc = *((int*) ifInDiscards);
    if (ifOutErrors)
	counters[ix].oerrs = *((int*) ifOutErrors);
    if (ifOutDiscards)
	counters[ix].odisc =  *((int*) ifOutDiscards);
    if (ifOutQLen)
	counters[ix].oqlen =  *((int*) ifOutQLen);
    
    if (print_hook)
	(*print_hook) (ix, ifDescr);

    for (vip = vars; vip; vip = vip -> next)
	(*vip -> os -> os_free) (vip -> value);
}


void
B_interval (interval, ps, sd)
    int interval;
    PS ps;
    int sd;
{
    int first = 1;
    int oldmask;
    struct type_SNMP_Message *msg;

    while(1)
    {
	if (!hflag && !first)
	    Bint_heading();

	msg = new_message (type_SNMP_PDUs_get__next__request, community);
	if (!msg)
	    adios (NULLCP, "out of memory");
    
	oid_ifIndex       = add_null_varbind (msg, "ifIndex");
	oid_ifDescr       = add_null_varbind (msg, "ifDescr");
	oid_ifInOctets    = add_null_varbind (msg, "ifInOctets");
	oid_ifInUcastPkts = add_null_varbind (msg, "ifInUcastPkts");
	oid_ifInNUcastPkts= add_null_varbind (msg, "ifInNUcastPkts");
	oid_ifOutOctets   = add_null_varbind (msg, "ifOutOctets");
	oid_ifOutUcastPkts= add_null_varbind (msg, "ifOutUcastPkts");
	oid_ifOutNUcastPkts=add_null_varbind (msg, "ifOutNUcastPkts");
	
	(void) signal (SIGALRM, catchalarm);
	signalled = 0;
	(void) alarm (interval);

	if (first)
	{
	    counters = old_counters;
	    getnext_process (msg, ps, sd, process_int);
	    counters = new_counters;
	    print_hook = print_B_int;
	}
	else
	    getnext_process (msg, ps, sd, process_int);
	
	if (!first)
	    bcopy((char*) new_counters, (char*) old_counters,
		  sizeof (struct if_counters) * (max_if_num+1));

	first = 0;

	oldmask = sigblock (sigmask (SIGALRM));
	if (! signalled) {
	    sigpause (0);
	}
	sigsetmask (oldmask);
	signalled = 0;
    }
}

void
C_interval (interval, ps, sd)
    int interval;
    PS ps;
    int sd;
{
    int first = 1;
    int oldmask;
    struct type_SNMP_Message *msg;

    while(1)
    {
	if (!hflag && !first)
	    Cint_heading();

	msg = new_message (type_SNMP_PDUs_get__next__request, community);
	if (!msg)
	    adios (NULLCP, "out of memory");
    
	oid_ifIndex       = add_null_varbind (msg, "ifIndex");
	oid_ifDescr       = add_null_varbind (msg, "ifDescr");
	oid_ifInOctets    = add_null_varbind (msg, "ifInOctets");
	oid_ifInUcastPkts = add_null_varbind (msg, "ifInUcastPkts");
	oid_ifInNUcastPkts= add_null_varbind (msg, "ifInNUcastPkts");
	oid_ifInErrors    = add_null_varbind (msg, "ifInErrors");
	oid_ifInDiscards  = add_null_varbind (msg, "ifInDiscards");
	oid_ifOutOctets   = add_null_varbind (msg, "ifOutOctets");
	oid_ifOutUcastPkts= add_null_varbind (msg, "ifOutUcastPkts");
	oid_ifOutNUcastPkts=add_null_varbind (msg, "ifOutNUcastPkts");
	oid_ifOutErrors   = add_null_varbind (msg, "ifOutErrors");
	oid_ifOutDiscards = add_null_varbind (msg, "ifOutDiscards");
	oid_ifOutQLen     = add_null_varbind (msg, "ifOutQLen");
	
	(void) signal (SIGALRM, catchalarm);
	signalled = 0;
	(void) alarm (interval);

	if (first)
	{
	    counters = old_counters;
	    getnext_process (msg, ps, sd, process_int);
	    counters = new_counters;
	    print_hook = print_C_int;
	}
	else
	    getnext_process (msg, ps, sd, process_int);

	if (!first)
	    bcopy((char*) new_counters, (char*) old_counters,
		  sizeof (struct if_counters) * (max_if_num+1));

	first = 0;
	
	oldmask = sigblock (sigmask (SIGALRM));
	if (! signalled) {
	    sigpause (0);
	}
	sigsetmask (oldmask);
	signalled = 0;
    }
}


void
intpr (interval, ps, sd)
    int interval;
    PS ps;
    int sd;
{
    struct type_SNMP_Message *msg;
    
    get_ip_addr_info (ps, sd);

    msg = new_message (type_SNMP_PDUs_get__next__request, community);
    if (!msg)
	adios (NULLCP, "out of memory");
	
    if (interval) 
    {
	max_if_num = get_int_var (ps, sd, "ifNumber.0");
	new_counters =
	    (struct if_counters*)
		calloc (sizeof (struct if_counters), max_if_num+1);
	old_counters = 
	    (struct if_counters*)
		calloc (sizeof (struct if_counters), max_if_num+1);
    }

    if (iflag && !Cflag && !Bflag) 
    {
	if (!hflag) 
	    iheading ();

	oid_ifIndex       = add_null_varbind (msg, "ifIndex");
	oid_ifDescr       = add_null_varbind (msg, "ifDescr");
	oid_ifMtu         = add_null_varbind (msg, "ifMtu");
	oid_ifAdminStatus = add_null_varbind (msg, "ifAdminStatus");
	oid_ifOperStatus  = add_null_varbind (msg, "ifOperStatus");
	oid_ifType        = add_null_varbind (msg, "ifType");
	oid_ifPhysAddress = add_null_varbind (msg, "ifPhysAddress");
	oid_ifSpeed       = add_null_varbind (msg, "ifSpeed");

	getnext_process (msg, ps, sd, process_if_info);
	return;
    }
    if (Bflag)
    {
	if (interval) 
	    B_interval(interval, ps, sd);
	else
	{
	    if (!hflag)
		Bheading();
	    
	    oid_ifIndex       = add_null_varbind (msg, "ifIndex");
	    oid_ifDescr       = add_null_varbind (msg, "ifDescr");
	    oid_ifInOctets    = add_null_varbind (msg, "ifInOctets");
	    oid_ifInUcastPkts = add_null_varbind (msg, "ifInUcastPkts");
	    oid_ifInNUcastPkts= add_null_varbind (msg, "ifInNUcastPkts");
	    oid_ifOutOctets   = add_null_varbind (msg, "ifOutOctets");
	    oid_ifOutUcastPkts= add_null_varbind (msg, "ifOutUcastPkts");
	    oid_ifOutNUcastPkts=add_null_varbind (msg, "ifOutNUcastPkts");

	    getnext_process (msg, ps, sd, process_B_info);
	}
    }
    if (Cflag)
    {
	if (interval)
	{
	    C_interval(interval, ps, sd);
	}
	else
	{
	    if (!hflag)
		Cheading();
	    
	    oid_ifIndex       = add_null_varbind (msg, "ifIndex");
	    oid_ifDescr       = add_null_varbind (msg, "ifDescr");
	    oid_ifInUcastPkts = add_null_varbind (msg, "ifInUcastPkts");
	    oid_ifInNUcastPkts= add_null_varbind (msg, "ifInNUcastPkts");
	    oid_ifInErrors    = add_null_varbind (msg, "ifInErrors");
	    oid_ifInDiscards  = add_null_varbind (msg, "ifInDiscards");
	    oid_ifOutUcastPkts= add_null_varbind (msg, "ifOutUcastPkts");
	    oid_ifOutNUcastPkts=add_null_varbind (msg, "ifOutNUcastPkts");
	    oid_ifOutErrors   = add_null_varbind (msg, "ifOutErrors");
	    oid_ifOutDiscards = add_null_varbind (msg, "ifOutDiscards");

	    getnext_process (msg, ps, sd, process_C_info);
	}
    }
}

static void
catchalarm()
{
    signalled = 1;
}
